package com.datarp.calc;

import java.util.*;


/**

The Calculator class parses and computes the value of arithmetic 
expressions using the four basic arithmentic opertors and an unlimited
number of nested parenthesis. The computations are performed using integer
operations only to facilitate floating point operations in environments
where the primitive types float and double are unavailable.<P>

For example, the Java expression,
	<code>new Calculator("(2.01+3.5)*2.2").evaluate()</code>
will return the String "12.122".<P>

The Calculator class uses the same order of operations as the Java(TM) language
for the operators it supports:
<UL><LI>Level 0: literals
<LI>	Level 1: ( )
<LI>	Level 2: unary -, +
<LI>	Level 3: * /
<LI>	Level 4: + -
</UL>
Numeric literals are composed of a positive or negative floating point base
optionally followed by the letter e or E and a positive or negative exponent.<P>

Examples of valid literals are:
<UL><LI>1
<LI>	-1
<LI>	2.5
<LI>	-3.03
<LI>	1e10
<LI>	1E-10
<LI>	123.456789e-1234
</UL>
It is critical to understand the role of precision in performing computations. 
The primitive data type long is used to hold the base of floating point numbers.
As such, there is a finite amount of precision that can be stored internally. Further,
a given computation may require additional precision, which limits the amount of initial
precision that is available. By default, the precision is set to be half the number of 
digits found in Long.MAX_VALUE. In standard Java, this is nine digits of precision.
All numbers read in will be rounded to the set number of digits of precision, any additional
powers of ten are transfered to the exponent. Likewise, all results are rounded to
the set number of digits of precision.<P>

The default choice of 9 digits of precision gaurantees that the largest integer that 
can be represented in nine digits times itself will not produce an integer overflow condition.
This same claim can be be made for any precision less than nine, AND CANNOT be made for
any precision greater than nine.<P>

The default is a good choice for many computations, but NOT for all computations. For example,
the sum <code>1e10 + 1</code> produces the answer <code>1e10</code> if
nine digits of precision are used. Increasing the precision to eleven digits produces
the answer <code>10000000001</code>.<P>

As with any floating point computation, it is important to fully understand the manner
in which any calculation is made, and to extensively test results for accuracy. The
Calculator class is incapable of performing computations with high levels of precision.<P>

The Calculator class explicitly does not perform computations in the same manner
as the JVM performs float and double primitive computations, and is not gauranteed
to produce identical results, or to round results by any common convention. Further,
Data Representations, Inc. expressly states that any results produced by the Calculator
class may be entirely inaccurate or incorrect, and Data Representations, Inc.
disclaims any liability that arises from the use of such results. 
Data Representations expressly states that the Calculator class is not fit
for any purpose.<P>
	
 *
 * @author  Carl H. Sayres
 * @author  Data Representations, Inc.
 * @version 1.0, 10/29/00
 */
public class Calculator {

	private static final char[] plus = new char[]{'+'};
	private static final char[] minus = new char[]{'-'};
	private static final char[] multiply = new char[]{'*'};
	private static final char[] divide = new char[]{'/'};
	private static final char[] open = new char[]{'('};
	private static final char[] close = new char[]{')'};

	private static int MAX_PRECISION;
	static {
		StringBuffer sb = new StringBuffer();
		sb.append(Long.MAX_VALUE);
		MAX_PRECISION = sb.length();
		sb.setLength(0);
		sb.append(Long.MIN_VALUE);
		if ( sb.length()-1 < MAX_PRECISION ) MAX_PRECISION = sb.length()-1;
		MAX_PRECISION--;
	}

	private Stack stack = new Stack();
	private Cpu cpu = new Cpu();
	private Lexer lexer;
	private char[] token = null;

	private String expression;
	private int precision;

	public static void main(String args[]) {
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<args.length; i++) {
			sb.append(args[i]);
			sb.append(' ');
		}
		Calculator calc = new Calculator(sb.toString());
		for ( int i=2; i<18; i++ ) {
			calc.setPrecision(i);
			try {
				System.out.println(i+": "+calc.evaluate());
			} catch ( Exception e ) {
				System.out.println(i+": "+e.getClass().getName()+": "+e.getMessage());
			}
		}
	}

	public Calculator(String expression) {
		this(expression,MAX_PRECISION/2);
	}

	public Calculator(String expression, int precision) {
		setExpression(expression);
		setPrecision(precision);
	}

	public void setExpression(String expression) {
		if ( expression == null ) throw new NullPointerException("Expression cannot be null");
		this.expression = expression;
	}

	public void setPrecision(int precision) {
		if ( precision < 2 ) throw new IllegalArgumentException("Precision must be greater than 1");
		if ( precision > MAX_PRECISION ) throw new IllegalArgumentException("Precision must be less than "+MAX_PRECISION);
		this.precision = precision;
	}

	public String getExpression() {
		return expression;
	}

	public int getPrecision() {
		return precision;
	}

	public String evaluate() {
		stack.removeAllElements();
		lexer = new Lexer(expression);
		token = null;
		try {
			nextToken();
			parseLevel4();
			return new String((char[])stack.pop());
		} catch ( EmptyStackException e ) {
			throw new IllegalArgumentException(
			"Empty Stack Error at position "+lexer.getLastPosition());
		}
	}

	private final void nextToken() {
		token = lexer.getNextToken();
	}

	private void parseLevel4() throws EmptyStackException {
		parseLevel3();
		while ( token == plus || token == minus ) {
			char[] op = token;
			nextToken();
			parseLevel3();
			char[] operand2 = (char[])stack.pop();
			char[] operand1 = (char[])stack.pop();
			if ( op == plus ) {
				stack.push(cpu.setOp(operand1,operand2).add());
			} else {
				stack.push(cpu.setOp(operand1,operand2).subtract());
			}
		}
	}

	private void parseLevel3() throws EmptyStackException {
		parseLevel2();
		while ( token == multiply || token == divide ) {
			char[] op = token;
			nextToken();
			parseLevel2();
			char[] operand2 = (char[])stack.pop();
			char[] operand1 = (char[])stack.pop();
			if ( op == multiply ) {
				stack.push(cpu.setOp(operand1,operand2).multiply());
			} else {
				stack.push(cpu.setOp(operand1,operand2).divide());
			}
		}
	}

	private void parseLevel2() throws EmptyStackException {
		if ( token == plus || token == minus ) {
			char[] op = token;
			nextToken();
			parseLevel1();
			if ( op == minus ) {
				stack.push(cpu.setOp((char[])stack.pop()).negate());
			}
		}
		else parseLevel1();
	}

	private void parseLevel1() throws EmptyStackException {
		if ( token == open ) {
			nextToken();		// (
			parseLevel4();
			if ( token != close )
				throw new IllegalArgumentException(
				"Missing ')' at position "+lexer.getLastPosition());
			nextToken();		// )
		}
		else parseLevel0();
	}

	private void parseLevel0() throws EmptyStackException {
		if ( token == null ) {
			throw new IllegalArgumentException(
			"Incomplete Expression at position "+lexer.getLastPosition());
		} else if ( (token[0] >= '0' && token[0] <= '9') || token[0] == '.' ) {
			stack.push(token);
			nextToken();
		} else {
			throw new IllegalArgumentException(
			"Unexpected token at position "+lexer.getLastPosition()+": "+new String(token));
		}
	}

	class Lexer {
		private char[] buffer;
		private int cursor;
		private int last;

		Lexer(String expression) {
			buffer = expression.toCharArray();
			cursor = last = 0;
		}

		boolean hasMoreTokens() {
			return cursor < buffer.length;
		}

		int getLastPosition() {
			return last;
		}

		char[] getNextToken() {
			try {
				while (
					buffer[cursor] == ' ' ||
					buffer[cursor] == '\t' ||
					buffer[cursor] == '\n' ||
					buffer[cursor] == '\r'
				) cursor++;
				last = cursor;
				switch ( buffer[cursor] ) {
				case '+': cursor++; return plus;
				case '-': cursor++; return minus;
				case '*': cursor++; return multiply;
				case '/': cursor++; return divide;
				case '(': cursor++; return open;
				case ')': cursor++; return close;

				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case '.':
					int begin = cursor;
					try {
						while ( buffer[cursor] >= '0' && buffer[cursor] <= '9' ) cursor++;
						if ( buffer[cursor] == '.' ) {
							cursor++;
							while ( buffer[cursor] >= '0' && buffer[cursor] <= '9' ) cursor++;
						}
						if ( buffer[cursor] == 'e' || buffer[cursor] == 'E' ) {
							buffer[cursor] = 'e'; // to lower case
							cursor++;
							if ( buffer[cursor] == '+' || buffer[cursor] == '-') cursor++;
							while ( buffer[cursor] >= '0' && buffer[cursor] <= '9' ) cursor++;
						}
					} catch ( ArrayIndexOutOfBoundsException e ) {}
					char[] charar = new char[cursor-begin];
					System.arraycopy(buffer,begin,charar,0,cursor-begin);
					return charar;

				default:
					throw new IllegalArgumentException(
					"Illegal character at position "+cursor+": "+buffer[cursor]);
				}
			} catch ( ArrayIndexOutOfBoundsException e ) { return null; }
		}
	}

	class Cpu {
		long base1;
		long power1;
		long base2;
		long power2;

		Cpu setOp(char[] op1, char[] op2) {
			decodeInto1(op2);
			base2 = base1;
			power2 = power1;
			decodeInto1(op1);
			return this;
		}

		Cpu setOp(char[] op1) {
			decodeInto1(op1);
			return this;
		}

		final void decodeInto1(char[] op) {
			base1 = 0L;
			power1 = 0L;
			boolean count = false;
			int precisionleft = precision;
			for ( int i=(op[0]=='-'?1:0); i<op.length; i++ ) {
				if ( op[i] >= '0' && op[i] <= '9' ) {
					if ( precisionleft < 1 ) {
						if ( precisionleft == 0 && op[i] > '4' ) base1++; // round the base
						if ( !count ) power1++;
					} else {
						if ( count ) power1--;
						base1 = base1*10 + (op[i]-'0');
					}
					precisionleft--;
				} else if ( op[i] == '.' ) {
					count = true;
				} else if ( op[i] == 'e' ) {
					i++;
					boolean negate = op[i] == '-';
					if ( op[i] == '-' || op[i] == '+' ) i++;
					long base = 0;
					for ( ; i<op.length; i++ ) base = base*10 + (op[i]-'0');
					if ( negate ) base *= -1;
					power1 += base;
					break;
				}
			}
			if ( op[0] == '-' ) base1 *= -1;
			if ( base1 == 0L ) power1 = 0L;
		}

		final void balancePowers() {
			if ( power1 > power2 ) {
				if ( base1 < 0 ) {
					while ( power1 != power2 && base1*10 > (Long.MIN_VALUE/10) ) {
						base1 *= 10;
						power1--;
					}
				}
				else if ( base1 > 0 ) {
					while ( power1 != power2 && base1*10 < (Long.MAX_VALUE/10) ) {
						base1 *= 10;
						power1--;
					}
				}
				while ( power1 != power2 ) {
					base2 /= 10;
					power2++;
				}
			}
			if ( power2 > power1 ) {
				if ( base2 < 0 ) {
					while ( power1 != power2 && base2*10 > (Long.MIN_VALUE/10) ) {
						base2 *= 10;
						power2--;
					}
				}
				else if ( base2 > 0 ) {
					while ( power1 != power2 && base2*10 < (Long.MAX_VALUE/10) ) {
						base2 *= 10;
						power2--;
					}
				}
				while ( power1 != power2 ) {
					base1 /= 10;
					power1++;
				}
			}
		}

		char[] add() {
			balancePowers();
			return format(base1 + base2, power1);
		}

		char[] subtract() {
			balancePowers();
			return format(base1 - base2, power1);
		}

		char[] multiply() {
			return format(base1 * base2, power1 + power2);
		}

		char[] divide() {
			if ( base1 < 0 ) {
				while ( base1*10 > (Long.MIN_VALUE/10) ) {
					base1 *= 10;
					power1--;
				}
			}
			else if ( base1 > 0 ) {
				while ( base1*10 < (Long.MAX_VALUE/10) ) {
					base1 *= 10;
					power1--;
				}
			}
			return format(base1 / base2, power1 - power2);
		}

		char[] negate() {
			return format(-1*base1,power1);
		}

		char[] format(long base, long power) {
			if ( base == 0L ) power = 0L;
			StringBuffer sb = new StringBuffer();
			boolean neg = base < 0;
			if ( neg ) base *= -1;
			sb.append(base);
			if ( sb.length() > precision ) {
				boolean round = sb.charAt(precision) > '4';
				power += sb.length()-precision;
				sb.setLength(precision);
				if ( round ) {
					base = 0;
					for ( int i=0; i<sb.length(); i++)
						base = base*10 + (sb.charAt(i)-'0');
					base++;
					sb.setLength(0);
					sb.append(base);
				}
			}
			if ( power < 0 ) {
				if ( -1*power < sb.length() ) {
					sb.insert( (int)(sb.length()+power), '.' );
					power = 0;
				} else if ( -1*power == sb.length() ) {
					sb.insert( 0, '0' );
					sb.insert( 1, '.' );
					power = 0;
				} else {
					power += sb.length()-1;
					sb.insert( 1, '.' );
				}
				while ( sb.charAt(sb.length()-1) == '0' ) sb.setLength(sb.length()-1);
				if ( sb.charAt(sb.length()-1) == '.' ) sb.setLength(sb.length()-1);
			}
			else if ( power > 0 ) {
				if ( sb.length()+power <= precision ) {
					while ( power > 0 ) {
						sb.append('0');
						power--;
					}
				} else {
					power += sb.length()-1;
					sb.insert( 1, '.' );
					while ( sb.charAt(sb.length()-1) == '0' ) sb.setLength(sb.length()-1);
					if ( sb.charAt(sb.length()-1) == '.' ) sb.setLength(sb.length()-1);
				}
			}
			if ( power != 0 ) {
				sb.append('e');
				sb.append(power);
			}
			if ( neg ) sb.insert( 0, '-' );
			char[] output = new char[sb.length()];
			sb.getChars(0,output.length,output,0);
			return output;
		}

	}
}

